/**/

   /*  WhatWPIpkg.cmd                                                                                 */
   /*  Anton Monroe <amonroe5301@gmail.com>                                                           */
   /*                                                                                                 */
   /*  reports which WarpIn package a file came from                                                  */
   /*  this is normally called from whatowns.btm, but it will also work on its own                    */

   Main: /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   call on Error
   signal on NoValue

   parse arg Mainargs

   call Init
   Opt. = ''
   ValidSwitch = 'help terse'
   interpret GetOpts(Mainargs)
   if Opt.0help <> '' then do
         call Usage
         return 0
      end
   if Opt.0errors > 0 then
      /*  the message was already displayed when GetOpts() was interpreted                            */
      call Egress 'Usage'

   select
      when ArgC > 1 then do
            call lineout 'stderr', G.0meID 'too many arguments'
            call Usage
            call Egress 'usage'
         end

      when ArgC < 1 then do
            call lineout 'stderr', G.0meID 'not enough arguments'
            call Usage
            call Egress 'usage'
         end
      otherwise
         G.0Infile = ArgV.1
         G.0Infile = translate(G.0Infile, '\', '/')
   end   /*  select  */

   /* find out where the WarpIn database is                                                           */
   dummy. = ''
   inifile = 'datbas_'||left(SysBootDrive(), 1)||'.ini'
   WIpath = strip(SysIni(, 'WarpIN', 'Path'), , '00'x)
   inifile = WIpath||'\'||inifile
   if SysFileTree(inifile, 'dummy', 'FO') <> 0 then do
         call lineout 'stderr', G.0meID 'cannot find Warpin database 'inifile
         call Egress 'error'
      end
   drop dummy.

   /*  Use stream() to get the full path of the file.  But it only works for a file that exists.      */
   FullSpec = stream(G.0Infile, 'c', 'query exists')
   if FullSpec == '' then do
         if substr(G.0Infile, 2, 2) <> ':\' then do
               call lineout 'stderr', G.0meID 'Could not resolve the filespec "'G.0Infile'"'
               call lineout 'stderr', '         try giving the complete path'
               call Egress 'usage'
            end
         else do
               /*  G.0Infile has full path but file not found. Maybe a missing file that              */
               /*  is still in the WarpIn database                                                    */
               FullSpec = G.0Infile
            end
      end

   Basepath = filespec('drive', FullSpec)||filespec('path', FullSpec)
   Basepath = translate(Basepath)
   /* get a list of files that match Fullspec                                                         */
   call SysFileTree FullSpec, 'FileN', 'FO', '*****'
   if FileN.0 == 0 then do
         /*  if no match, use FullSpec; this lets us check for a non-existent file                    */
         FileN.1 = FullSpec
         FileN.0 = 1
      end

   /* search the INI file                                                                             */
   Apps. = ''
   call SysIni inifile, 'All:', 'Apps'
   if Result == 'ERROR:' then do
         call lineout 'stderr',  G.0meID 'Error reading 'inifile
         call lineout 'stderr',  G.0meID 'Maybe because WarpIn is running?'
         call Egress 'error'
      end

   WI. = 0
   found = 0
   p = 0
   do i = 1 to Apps.0
      AllFilesz = ''
      pathz = SysIni(inifile, Apps.i, 'TargetPath')
      parse var pathz path '00'x
      path = translate(path)||'\'
      comp = compare(path, Basepath)
      if comp == length(path) +1 ,
       | comp = length(Basepath) + 1 ,
       | comp == 0 then do
            /* the beginning of both paths is the same; there might be some matching files            */
            p = p + 1
            WI.0 = p
            WIpath.p = path
            parse var Apps.i . '\' pkg '\' subpkg '\' .
            WIpkg.p = pkg' ('subpkg')'
            call SysIni inifile, Apps.i, 'All:', 'Keys'
            if Result == 'ERROR:' then call Error
            do j = 1 to Keys.0
               /* 'Files' sometimes overflows into 'Files02' so read all of the 'Filesxxx'            */
               if left(Keys.j, 5) == 'Files' then do
                     filesz = SysIni(inifile, Apps.i, Keys.j)
                     if Result == 'ERROR:' then call Error
                     AllFilesz = AllFilesz||filesz
                  end
            end j
            WIfiles.p = translate(AllFilesz)
         end
   end i

   do i = 1 to WI.0
      FileU = ''
      comp = 0
      do j = 1 to FileN.0
         comp = compare(WIpath.i, translate(FileN.j))
         if comp == length(WIpath.i) +1 then do
               /*
               the file we are checking is under the install dir of this package
               The database stores filespecs relative to the install dir, so trim the
               install dir from our filespec and see if what is left matches a filename
               in the database.
               There is no point in parsing the file list, all we care about is if the
               filename is there.
               */
               FileU = translate(substr(FileN.j, comp))
               if pos(FileU, WIfiles.i) > 0 then do
                     found = found + 1
                     if Opt.0terse == '' then do
                           say '  'FileN.j
                           say "        belongs to  '"WIpkg.i"'"
                        end
                     else say WIpkg.i
                  end
            end
      end j
   end i
   if Opt.0terse == '' then
      say 'Found 'found' matches'
   call Egress 'okay'
   /* === End of Main =============================================================================== */

   Init: /*fold00*/
   /* =============================================================================================== */
   /*  load DLLs, initialize the G. stem of constants. Add anything else I want                       */
   /* =============================================================================================== */
   if RxFuncQuery('SysLoadFuncs') then do
         call RxFuncAdd 'SysLoadFuncs', 'REXXUTIL', 'SysLoadFuncs'
         call SysLoadFuncs
      end
   if RxFuncQuery('RxQprocstatus') then do
         call RxFuncAdd 'RxuInit', 'RXU', 'RxuInit'
         call RxuInit
      end

   /* My collection of external functions                                                             */
   /*     call SysLoadRexxMacroSpace 'F:\akm\btm\rexx\macro\akmlib.macro'                             */

   signal on Syntax
   signal on NoValue

   G. = ''
   /* ----------------------------------------------------------------------------------------------- */
   /*  G.0me' is the name of this file. (When a Rexx script is invoked from the                       */
   /*  command line, the name is in ALL CAPS, which is hard to read.  To get the                      */
   /*  correct capitalization of it, the only way I have found is a cumbersome call to                */
   /*  SysFileTree(), and even that only corrects the filename, not the path.)                        */
   parse source . . G.0me
   call SysFileTree G.0me, 'tmp.', 'FO'
   G.0me_long = tmp.1
   G.0me = filespec('name', G.0me_long)
   G.0meID = '['G.0me']'
   G.0me_ind = copies(' ', length(G.0me))
   G.0meID_ind = copies(' ', length(G.0meID))
   drop tmp.
   /* ----------------------------------------------------------------------------------------------- */
   Shared_vars = 'sigl Opt. G.'
   return 0
   /* === End of Init =============================================================================== */

   Usage: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /* =============================================================================================== */
   call lineout 'stderr', ''
   call lineout 'stderr', G.0me 'shows which WarpIn package owns a file'
   call lineout 'stderr', ''
   call lineout 'stderr', 'Usage:'
   call lineout 'stderr', '   'G.0me '<filespec> [option]'
   call lineout 'stderr', 'Options:'
   call lineout 'stderr', '    -terse        show only the package name'
   call lineout 'stderr', '    -help         show this help'
   call lineout 'stderr', ''
   call lineout 'stderr', 'Examples:'
   call lineout 'stderr', '   'G.0me 'Path\to\file.txt'
   call lineout 'stderr', '   'G.0me 'Path\to\file.txt -terse'
   call lineout 'stderr', '   'G.0me 'Path\to\*'
   call lineout 'stderr', ''
   return 1
   /* === End of Usage ============================================================================== */

   ShowGetOpts: procedure expose (Shared_vars) ArgC ArgV. /*fold00*/
   0ShowGetOpts:
   /* =============================================================================================== */
   /*    show what GetOpts() returned and how it was interpreted                                      */
   /* =============================================================================================== */
   say copies('=', 80)
   say ' GetOpts returned these arguments:'
   do i = 1 to ArgC
      say '   ' ArgV.i
   end
   say copies('=', 80)

   /*    show the Opt.* variables                                                                     */
   VarList = 'Opt.0optlist 'Opt.0optlist' ArgC'
   do i = 0 to ArgC
      VarList = VarList' ArgV.'i
   end
   do while VarList <> ''
      parse var VarList thisvar VarList
      say '  'overlay('=  'value(thisvar), thisvar, 25)
   end
   return 0
   /* === End of 0ShowGetOpts ======================================================================= */

   NoValue: /*fold00*/
   0NoValue:
   /* =============================================================================================== */
   /*  signal on NoValue                                                                              */
   /* ----------------------------------------------------------------------------------------------- */
   /*  depends on variables: G.0meID sigl                                                             */
   /*  depends on functions: Egress()                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   signal off NoValue
   _ = sigl
   call lineout 'stderr', G.0meID'[NOVALUE Handler] at line '_' for variable "'condition('d')'"'
   call lineout 'stderr', '   ==> {'sourceline(_)'}'
   call Egress 'bug_NoValue'
   /* === End of 0NoValue =========================================================================== */

   Mark: /*fold00*/
   0Mark:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       CALL Mark [<list of variable names>] [, <string>]                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Shows the filename and line number it was called from, the current values of the               */
   /*  variables named, and an informational string                                                   */
   /*  Arg(1) is a list of variable names                                                             */
   /*  Arg(2) can be any descriptive text                                                             */
   /*  Examples:                                                                                      */
   /*      CALL mark 'varA varB', 'At top of the "DO x = 1 to 5" loop'                                */
   /*  If no variable names are given, it will just show a message:                                   */
   /*      CALL mark '', 'At top of the "DO x = 1 to 5" loop'                                         */
   /*                                                                                                 */
   /*  This must share all the variables of whatever routine it was called from.                      */
   /*  I cannot think of a way to do it without setting at least one variable here,                   */
   /*  so reserve the variable '_' for the use of Mark and similar routines.                          */
   /* =============================================================================================== */
   _ = sigl
   call lineout 'stderr', ''
   if arg(2, 'e') ,
    & arg(2) = '' then nop
   else call lineout 'stderr', '[Mark]['value('G.0me')':'_']' arg(2)
   do _ = 1 to 99
      if word(arg(1), _) = '' then leave _
      if symbol(word(arg(1), _)) = 'VAR' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||'= "'||value(word(arg(1), _))||'"'
      if symbol(word(arg(1), _)) = 'LIT' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not defined'
      if symbol(word(arg(1), _)) = 'BAD' then
         call lineout 'stderr', '      '||left(word(arg(1), _), 30)||' is not a variable'
   end _
   drop _
   return 0
   /* === End of 0Mark ============================================================================== */

   LineNo: /*fold00*/
   0LineNo:
   /* =============================================================================================== */
   /*  returns the number of the current line--    say 'now at line 'LineNo()                         */
   /* =============================================================================================== */
   _ = sigl
   return _
   /* === End of 0LineNo ============================================================================ */

   LineID: /*fold00*/
   0LineID:
   /* =============================================================================================== */
   /*    LineID()                                                                                     */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns a string identifying this file and the line number it was called from                  */
   _ = sigl
   return '['G.0me':'right(_, 5, '0')']'
   /* === End of 0LineID ============================================================================ */

   ShowError: procedure expose (Shared_vars) /*fold00*/
   0ShowError:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        call ShowError <Arg1> [, <Arg2>]                                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*   sends a message to stderr, adding a header with the program name, function name,              */
   /*   and line number to show where the message comes from.                                         */
   /*   Use '' as the second arg to improve the display of a multi-line message:                      */
   /*                                                                                                 */
   /*   Example: in function Add, to show a multi-line message:                                       */
   /*       call ShowError 'Warning: unexpected value for variable varA: "'varA'"'                    */
   /*       call ShowError 'It should be a number', ''                                                */
   /*   displays:                                                                                     */
   /*       [Foo.cmd:Add:33] Warning: unexpected value for variable varA: "foo"                       */
   /*                        It should be a number                                                    */
   /*                                                                                                 */
   /* =============================================================================================== */
   seSigl = sigl

   if Opt.0verbose \= 0 then
      seID = '['G.0me':'LastLabel(seSigl)':'seSigl']'
   else
      seID = '['G.0me':'seSigl']'

   if arg(2, 'e') ,
    & arg(2) = '' then
      seID = copies(' ', length(seID))
   else
      call lineout 'stderr', ''

   call lineout 'stderr', seID arg(1)
   return 0
   /* === End of 0ShowError ========================================================================= */

   Syntax: /*fold00*/
   0Syntax:
   /* =============================================================================================== */
   /*  depends on variables: G.0me sigl                                                               */
   /*  depends on functions: LastLabel(), ShowError(), Egress()                                       */
   /* ----------------------------------------------------------------------------------------------- */
   /*  traps syntax errors and shows a better error message than Rexx does. In particular,            */
   /*  an incomplete DO/SELECT/IF is fairly common and the error message from Rexx is no              */
   /*  help at all.                                                                                   */
   /*                                                                                                 */
   /*  Some errors cannot be handled by an error handler, like REX06 "unmatched comment or quote"     */
   /*  (Which makes sense. If a defective script cannot be tokenised, then it cannot be run, so       */
   /*  the handler never exists. But the help message for REX06 implies that error 6 may not be       */
   /*  detected until later, so there is something I don't understand.)                               */
   /*                                                                                                 */
   /*  This will always be a work in progress                                                         */
   /* =============================================================================================== */
   SyntaxSigl = sigl
   call lineout 'stderr', ''
   /*  sometimes a "control stack full" will be trapped here and sometimes not. I don't know why.     */
   /*  If the control stack is full, any DO or SELECT after this will probably trigger                */
   /*  another (untrapped) error 11. Get the important part of the message onto the                   */
   /*  screen before that happens, without using DO.                                                  */
   /*  A trivial bug: The default formatting style puts a space before a continuation                 */
   /*  comma. In this case the space becomes part of the string to be displayed, at                   */
   /*  the start of the next line. Not worth fixing.                                                  */
   if rc = 11 then
      call lineout 'stderr' ,
           '[SYNTAX Handler] Rexx error 11: Control Stack full at line' SyntaxSigl||G.0CRLF ,
           '                      ==>' strip(sourceline(SyntaxSigl))||G.0CRLF ,
           '                There may be another error 11 after this'||G.0CRLF
   else do
         select
            when rc == 14 then do
                  /*  incomplete DO/SELECT/IF                                                         */
                  /* Rexx reports an error at the end of file, which is useless                       */
                  call ShowError '[SYNTAX Handler]'
                  call ShowError '    Syntax error 'rc' which is "'errortext(rc)'"', ''
                  if SyntaxSigl >= sourceline() then do
                        call ShowError '    Rexx says it is at line 'SyntaxSigl', which is the end of file.', ''
                     end
                  else do
                        call ShowError '    at line 'SyntaxSigl, ''
                        call ShowError '       ==> 'strip(sourceline(SyntaxSigl)), ''
                     end
               end
            otherwise
               Err_func = LastLabel(SyntaxSigl)
               call ShowError '[SYNTAX Handler] Syntax error at line 'SyntaxSigl 'in label' Err_func
               call ShowError '==>' sourceline(SyntaxSigl), ''
               call ShowError 'Error # 'rc' is "'errortext(rc)'"', ''
               if condition('d') \= '' then call ShowError 'condition(D) is "'condition('d')'"', ''
         end   /*  select  */
      end
   call ShowError 'for more help, type "helpmsg REX'rc'"', ''
   call Egress 'BUG_Syntax'
   /* === End of 0Syntax ============================================================================ */

   Error: /*fold00*/
   0Error:
   /* =============================================================================================== */
   /*  signal on error                                                                                */
   /* ----------------------------------------------------------------------------------------------- */
   /*  this does not do much, but it is not used much anyway                                          */
   /* =============================================================================================== */
   signal off Error
   call lineout 'stderr', '[ErrorHandler] 'rc' at line 'sigl
   call lineout 'stderr', '       Error # 'rc' is "'errortext(rc)'"'
   call lineout 'stderr', '       for more help, type "helpmsg REX'rc'"'
   call Egress 'error'
   /* === End of 0Error ============================================================================= */

   LastLabel: procedure expose (Shared_vars) /*fold00*/
   0LastLabel:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*       ret = LastLabel()                                                                         */
   /*       ret = LastLabel(<line number>)                                                            */
   /* ----------------------------------------------------------------------------------------------- */
   /*  returns the name of the nearest label =in this program= previous to <line number>              */
   /*  if <line> is not given then it is the current line                                             */
   /*  returns '' if no label was found or if <line> is not a valid line number                       */
   /*                                                                                                 */
   /*  it is normally used to determine the name of a routine or function:                            */
   /*      Foo: procedure                                                                             */
   /*      say 'Now in procedure 'Lastlabel()         > "Now in procedure Foo"                        */
   /*                                                                                                 */
   /*  akm: This partly depends on the style of writing I use, where                                  */
   /*  multiline comments begin at the beginning of a line and end at the end of a                    */
   /*  line. It could be fooled by something that looks like a label inside other                     */
   /*  kinds of multiline comments. But it is useful, and if it is wrong it will only                 */
   /*  affect the progress and debugging messages. This is about as reliable as I can                 */
   /*  make it without parsing every comment letter by letter.                                        */
   /* =============================================================================================== */
   if arg(1, 'o') then sline = sigl ; else sline = arg(1)
   if datatype(sline, 'W') \= 1 then return ''
   if sline > sourceline() then return ''
   incomment = 0
   do l = sline to 1 by -1
      aline = strip(sourceline(l))
      if left(aline, 2) == '/'||'*' ,
       | right(aline, 2) == '*'||'/' then do
            /*  counter-intuitively, incomment is incremented by the end of comment,                  */
            /*  because we are working backwards                                                      */
            incomment = incomment - CountStr('/'||'*', aline) + CountStr('*'||'/', aline)
         end

      if incomment <= 0 then do
            if HasLabel(aline) then do
                  parse var aline label ':' .
                  label = strip(label)
                  return label
               end
         end
   end l
   return ''
   /* === End of 0LastLabel ========================================================================= */

   HasLabel: procedure expose (Shared_vars)
   0HasLabel:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        if HasLabel(<string>) then ...                                                           */
   /*  returns True if <string> is a line that looks like a label                                     */
   /*  the calling routine is responsible for identifying multi-line comments                         */
   /* =============================================================================================== */
   /*  it is a label if str is one word followed by ':', followed by '', 'procedure', or a comment    */
   parse arg str
   str = strip(str)
   if pos(':', str) == 0 then return 0
   if left(str, 2) == '/'||'*' then return 0
   parse var str label ':' rest
   rest = strip(rest)
   if words(label) \= 1 then return 0
   if symbol(label) == 'BAD' then return 0
   if rest \= '' then
      select
         when translate(word(rest, 1)) == 'PROCEDURE' then
            return 1
         when left(rest, 2) == '/'||'*' then
            return 1
         otherwise
            return 0
      end   /*  select  */
   return 1
   /* === End of 0HasLabel ========================================================================== */

   Egress: /*fold00*/
   0Egress:
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     call Egress <exit type>                                                                     */
   /* ----------------------------------------------------------------------------------------------- */
   /*  All exits come through here.                                                                   */
   /*  If you need to, say, clean up temporary files, do it from here                                 */
   /*                                                                                                 */
   /*  OKAY    == normal exit                                                                         */
   /*  USAGE   == user error                                                                          */
   /*  ERROR   == something else wrong                                                                */
   /*  BUG     == bug in the program                                                                  */
   /* =============================================================================================== */
   signal off NoValue
   signal off Syntax
   parse upper arg ExitCode

   /*    handle my typing mistakes                                                                    */
   if ExitCode = 'OKAY' then exit 0
   if ExitCode = '' then exit 0
   if ExitCode = 0 then exit 0
   if abbrev(ExitCode, 'HALT') then exit 1
   if abbrev(ExitCode, 'USAGE') then exit 1
   if abbrev(ExitCode, 'ERROR') then exit 3
   if abbrev(ExitCode, 'BUG') then exit 9
   /*    in case I used some other exit string and forgot to add it here:                             */
   exit 99
   /* === End of 0Egress ============================================================================ */

   CountStr: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        Count = CountStr(<needle>, <haystack>)                                                   */
   /*                                                                                                 */
   /*  counts how many times the string <needle> occurs in the string <haystack>, without overlap     */
   /*  if either <needle> or <haystack> is null, returns 0                                            */
   /*                                                                                                 */
   /*  this is an imitation of the CountStr() function from Object Rexx                               */
   /* =============================================================================================== */
   parse arg needle , haystack

   found = 0
   p = pos(needle, haystack)
   do while p > 0
      found = found +1
      p = pos(needle, haystack, p + length(needle))
   end
   return found
   /* === End of CountStr =========================================================================== */

   GetOpts: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     INTERPRET GetOpts(args [,valid])                                                            */
   /*                                                                                                 */
   /*  where 'args' is the command line to be parsed, and                                             */
   /*  'valid' is an optional space-separated list of valid switches                                  */
   /*  'help' and '?' are always valid                                                                */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     parse arg args                                                                              */
   /*     interpret GetOpts(args, 'switchA switchB')                                                  */
   /*                                                                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  Switches must start with '-'. For example:                                                     */
   /*                                                                                                 */
   /*         -Foo                (Opt.0Foo == 1)                                                     */
   /*         -Foo=1              (Opt.0Foo == 1)                                                     */
   /*         -!Foo or -Foo=0     (Opt.0Foo == 0)                                                     */
   /*         -Foo=string         (Opt.0Foo == 'string')                                              */
   /*         -Foo="string"       (Opt.0Foo == 'string')                                              */
   /*         -Foo=a,b,c          (Opt.0Foo == 'a,b,c')                                               */
   /*         -Foo="a, b, c"      (Opt.0Foo == 'a, b, c')                                             */
   /*         -Foo="a b c"        (Opt.0Foo == 'a b c')                                               */
   /*         -Foo=123            (Opt.0Foo == '123')                                                 */
   /*         -Foo= or -Foo=''    (Opt.0Foo == '')                                                    */
   /*                                                                                                 */
   /*  switches must not have spaces before or after the '='                                          */
   /*  values with spaces must be in quotes. the quotes will be stripped.                             */
   /*                                                                                                 */
   /*  a switch without a value is set to '1'                                                         */
   /*        -switchA                                                                                 */
   /*  is the same as                                                                                 */
   /*        -switchA=1                                                                               */
   /*                                                                                                 */
   /*  a switch can be negated with "!", so                                                           */
   /*        -!switchA                                                                                */
   /*  is the same as                                                                                 */
   /*        -switchA=0                                                                               */
   /*                                                                                                 */
   /*     -? is the equivalent of -help                                                               */
   /*     they will both result in:                                                                   */
   /*        Opt.0help    == 1                                                                        */
   /*                                                                                                 */
   /*  Interpreting what GetOpts returns will create several variables that the                       */
   /*  caller can use--                                                                               */
   /*     Opt.              is a stem variable that holds whatever was specified by switches          */
   /*     Opt.0optlist      a space-separated list of options that were read                          */
   /*     Opt.0errors       number of errors; that is, when GetOpts() did not understand something    */
   /*                                                                                                 */
   /*                       elements of the Opt. compound variable all start with '0', to avoid       */
   /*                       name conflicts with simple variables                                      */
   /*                                                                                                 */
   /*     ArgV.             is a stem that holds the non-switch arguments                             */
   /*                       (ArgV.0 is the name of the .cmd script, NOT the number of items           */
   /*                       in the stem)                                                              */
   /*     ArgC              is the number of non-switch arguments                                     */
   /*                                                                                                 */
   /*  Example:                                                                                       */
   /*     Foo.cmd -dir=up -msg="that way" -v "From here"                                              */
   /*     will result in these variables:                                                             */
   /*        Opt.0optlist == opt.0dir opt.0msg                                                        */
   /*        Opt.0dir     == up                                                                       */
   /*        Opt.0msg     == that way                                                                 */
   /*        Opt.0v       == 1                                                                        */
   /*        ArgV.0       == Foo.cmd                                                                  */
   /*        ArgV.1       == From here                                                                */
   /*        ArgC         == 1                                                                        */
   /*                                                                                                 */
   /*  in the case of errors, GetOpts() will set Opt.0errors to the number of errors                  */
   /*  found, and cause the caller to show a message about it on stderr. The caller                   */
   /*  should probably abort if Opt.0errors is greater than 0.                                        */
   /* =============================================================================================== */
   /* Some important variables (inside GetOpts) are:                                                  */
   /*    Str.SetOpts       == commands the the caller will execute to set options                     */
   /*                          in the Opt. stem variable                                              */
   /*    Str.SetArgs       == commands the the caller will execute to put the arguments               */
   /*                          in the ArgV. and ArgC variables                                        */
   /*    Str.Errors        == commands the caller will execute to report errors                       */
   /*                          reading the command line                                               */
   /*    ReturnStr         == all of the above, combined and returned to caller                       */
   /*    flag.             == variables for GetOpt's internal use                                     */
   /*                          flag.error == number of errors trying to read command line             */
   /*    Str.SwitchList    == list of options (switches) used on command line                         */
   /*    ArgC              == number of non-switch arguments                                          */
   /*    ArgV.             == array of non-switch arguments                                           */
   /*    valid             == list of valid switches                                                  */
   /*                                                                                                 */
   /* =============================================================================================== */
   flag. = ''
   flag.errors = 0
   str. = ''
   ReturnStr = ''
   ArgC = 0
   ArgV. = ''
   parse arg args , valid
   do i = 1 to WordsQ(args)
      thisArg = WordQ(args, i)
      if left(thisArg, 1) == '-' then
         call ReadSwitch thisArg
      else do
            /*    not a switch, add to Arg list                                                       */
            ArgC = ArgC +1
            ArgV.ArgC = thisArg
         end
   end
   /* ----------------------------------------------------------------------------------------------- */
   /*    now put it all together into a series of commands for the caller to interpret                */
   if flag.errors > 0 then do
         str.SwitchList = strip(str.SwitchList' Opt.0Errors')
      end
   str.SetOpts = str.SetOpts' Opt.0Errors = 'flag.errors ';'
   str.SetOpts = str.SetOpts 'Opt.0optlist = "'str.SwitchList'" ;'

   str.SetArgs = 'ArgV. = "" ; parse source . . ArgV.0 ;  ArgC = "'ArgC'" ;'
   do i = 1 to ArgC
      varVal = Unquoted(ArgV.i)
      str.SetArgs = str.SetArgs' ArgV.'i' = 'Quoted(varVal)' ;'
   end

   ReturnStr = str.errors str.SetOpts str.SetArgs
   return ReturnStr
   /* === End of GetOpts ============================================================================ */

   ReadSwitch: procedure expose flag. str. valid (Shared_vars)
   /* =============================================================================================== */
   /*  this is part of GetOpts()                                                                      */
   /* =============================================================================================== */
   parse arg thisArg
   Negate = 0
   switch = strip(thisArg, 'L', '-')
   if left(switch, 1) == '!' then do
         switch = strip(switch, 'L', '!')
         Negate = 1
      end
   parse var switch optName '=' optVal
   select
      /* -? and -help have fixed meanings and are allowed even if not included in 'valid'             */
      when translate(optName) == '?' then
         optName  = 'Opt.0Help'
      when translate(optName) == 'HELP' then
         optName  = 'Opt.0Help'
      when optName == '' then do
            str.errors = ,
              str.errors||'call lineout "stderr", "[GetOpts] `-` without a switch name" ;'
            flag.errors = flag.errors +1
            return
         end
      when valid \= '' ,
         & wordpos(translate(optName), translate(valid)) == 0 then do
            str.errors = str.errors|| ,
                         'call lineout "stderr", "[GetOpts] I do not understand switch `'thisArg'`" ;'
            flag.errors = flag.errors +1
            return
         end
      otherwise
         optName = 'Opt.0'optName
   end   /*  select  */

   select
      when Negate then
         if optVal == '' then
            optVal = 0
         else do
               str.errors = str.errors|| ,
                            'call lineout "stderr", "[GetOpts] I do not understand switch `'thisArg'`" ;'
               flag.errors = flag.errors +1
               return
            end
      when optVal == '' ,
         & right(switch, 1) \= '=' then
         /*  "-switch=" means Opt.0switch == '', not 1                                                */
         optVal = '1'
      otherwise
         /*    try to be tolerant of unusual arguments; only strip quotes if they are balanced        */
         /*    and only one level of nested quotes                                                    */
         optVal = Unquoted(optVal)
   end   /*  select  */

   if wordpos(optName, str.SetOpts) > 0 then do
         str.errors = ,
           str.errors||'call lineout "stderr", "[GetOpts] duplicate switch `'thisArg'`" ;'
         flag.errors = flag.errors +1
      end
   else do
         str.SetOpts = str.SetOpts' 'optName' = 'Quoted(optVal)' ;'
         str.SwitchList = strip(str.SwitchList' 'optName)
      end
   return
   /* === End of ReadSwitch ========================================================================= */

   Quoted: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Quoted(string)                                                                  */
   /*                                                                                                 */
   /*    if possible, quotes a string with the kind of quotes it does not contain                     */
   /*        {He said "no"}   --> {'He said "no"'}                                                    */
   /*        {He said 'no'}   --> {"He said 'no'"}                                                    */
   /*    otherwise uses double quotes and lets the caller deal with the errors                        */
   /* =============================================================================================== */
   parse arg string
   select
      when pos('"', string) == 0 then
         newstring = '"'string'"'
      when pos("'", string) == 0 then
         newstring = "'"string"'"
      otherwise
         newstring = '"'string'"'
   end
   return newstring
   /* === End of Quoted ============================================================================= */

   Unquoted: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*        string = Unquoted(string)                                                                */
   /*                                                                                                 */
   /*    strips quotes from a string if and only if it is enclosed in a balanced pair                 */
   /*    of single or double quotes                                                                   */
   /* =============================================================================================== */
   parse arg string
   if length(string) < 2 then return string
   select
      when left(string, 1) == '"' ,
         & right(string, 1) == '"' then
         string = substr(string, 2, length(string) -2)
      when left(string, 1) == "'" ,
         & right(string, 1) == "'" then
         string = substr(string, 2, length(string) -2)
      otherwise
         nop
   end
   return string
   /* === End of Unquoted =========================================================================== */

   WordQ: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordQ(<string>, <word number>, <separator characters>, <quote characters>)            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a replacement for the Rexx Word() function.                                                 */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Word() function--                                                                     */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg str , thisnum , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0
   Thisword  = ''
   w1        = 0
   w2        = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(str) ,
          & pos(substr(str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(str) then numwords = numwords +1
   if numwords == thisnum then w1 = i

   do while (i <= length(str))
      if pos(substr(str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(str) ,
                   & substr(str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(str, i, 1), chrsW) \= 0 ,
       & i <= length(str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            /* unless the end of the word we want, then quit                                          */
            if w1 > 0 then do
                  w2 = i -1
                  leave
               end
            do while (pos(substr(str, i, 1), chrsW) \= 0) ,
                   & (i <= length(str))
               i = i +1
            end
            if i <= length(str) then numwords = numwords +1
            /* if this is the start of the word we want, note position in string                      */
            if numwords == thisnum then w1 = i
         end   /*  wordsep  */
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   if i > length(str) then i = length(str)
   if w1 > 0 then do
         /* we know w1, if we ran out of string, w2 is the end of string                              */
         if w2 == 0 then w2 = i
         /* found start and end of word, copy it                                                      */
         wlen = w2 - w1 +1
         Thisword = substr(str, w1, wlen)
      end
   return Thisword
   /* === End of WordQ ============================================================================== */

   WordsQ: procedure expose (Shared_vars) /*fold00*/
   /* =============================================================================================== */
   /*  Usage:                                                                                         */
   /*     foo = WordsQ(string, <separator characters>, <quote characters>)                            */
   /*                                                                                                 */
   /*  The last two parameters are optional, so it will work                                          */
   /*  as a drop-in replacement for the Rexx Word() function.                                         */
   /* ----------------------------------------------------------------------------------------------- */
   /*  A better Words() function--                                                                    */
   /*  It recognizes a quoted string as one word                                                      */
   /*  Optional parameters let you decide which characters                                            */
   /*  count as word separators and as quotes.                                                        */
   /*                                                                                                 */
   /*  Default word separators are <space> and <tab>                                                  */
   /*  Default quote characters are single, double, and                                               */
   /*  back quotes-- ' " `                                                                            */
   /* =============================================================================================== */

   parse arg str , chrsW , chrsQ
   if chrsW == '' then chrsW = ' '||d2c(9)
   if chrsQ == '' then chrsQ = '"'||"'"||'`'
   i         = 1
   quote     = ''
   numwords  = 0

   /* find start of first word by trimming blanks and word characters                                 */
   do while i <= length(str) ,
          & pos(substr(str, i, 1), chrsW) > 0
      i = i +1
   end
   if i <= length(str) then numwords = numwords +1

   do while (i <= length(str))
      if pos(substr(str, i, 1), chrsQ) \= 0 then do
            /* start of quoted string;
             nothing counts until we get to the matching quote (or end of string) */
            quote = substr(str, i, 1)
            tsti = i +1
            /*  bugfix for "-opt=''"                                                                  */
            do while tsti < length(str) ,
                   & substr(str, tsti, 1) \= quote
               tsti = tsti +1
            end
            if substr(str, tsti, 1) == quote then
               /* found the matching quote                                                            */
               i = tsti
            else
               /* ran out of string with no closing quote, so ignore this quote                       */
               nop
            i = i +1
         end   /*  quote  */

      if pos(substr(str, i, 1), chrsW) \= 0 ,
       & i <= length(str) then do
            /* a word separator, skip all blanks, find start of next word                             */
            do while (pos(substr(str, i, 1), chrsW) \= 0) ,
                   & (i <= length(str))
               i = i +1
            end
            if i <= length(str) then numwords = numwords +1
         end
      else
         /* a normal character                                                                        */
         i = i +1
   end   /*  do i  */
   return numwords
   /* === End of WordsQ ============================================================================= */

